@@ -205,7 +205,7 @@ static struct pppcp_packet *pppcp_packet_new(struct pppcp_data *data,
 	guint16 packet_length = bufferlen + sizeof(*packet);
 
 	ppp_packet = ppp_packet_new(packet_length, data->driver->proto);
-	if (!ppp_packet)
+	if (ppp_packet == NULL)
 		return NULL;
 
 	/* advance past protocol to add CP header information */
@@ -310,9 +310,9 @@ static gboolean is_first_request(struct pppcp_timer_data *timer_data)
 
 /* actions */
 /* log an illegal event, but otherwise do nothing */
-static void pppcp_illegal_event(guint8 state, guint8 type)
+static void pppcp_illegal_event(GAtPPP *ppp, guint8 state, guint8 type)
 {
-	g_printerr("Illegal event %d while in state %d\n", type, state);
+	DBG(ppp, "Illegal event %d while in state %d", type, state);
 }
 
 static void pppcp_this_layer_up(struct pppcp_data *data)
@@ -690,7 +690,7 @@ static void pppcp_generate_event(struct pppcp_data *data,
 	return;
 
 error:
-	pppcp_illegal_event(data->state, event_type);
+	pppcp_illegal_event(data->ppp, data->state, event_type);
 }
 
 void pppcp_signal_open(struct pppcp_data *data)
@@ -938,6 +938,8 @@ void pppcp_send_protocol_reject(struct pppcp_data *data,
 
 	ppp_transmit(data->ppp, pppcp_to_ppp_packet(packet),
 			ntohs(packet->length));
+
+	pppcp_packet_free(packet);
 }
 
 /*
@@ -968,6 +970,8 @@ void pppcp_process_packet(gpointer priv, const guint8 *new_packet)
 
 void pppcp_free(struct pppcp_data *pppcp)
 {
+	pppcp_stop_timer(&pppcp->config_timer_data);
+	pppcp_stop_timer(&pppcp->terminate_timer_data);
 	g_free(pppcp->peer_options);
 	g_free(pppcp);
 }
@@ -994,22 +998,31 @@ void pppcp_set_local_options(struct pppcp_data *pppcp,
 	pppcp->local_options_len = len;
 }
 
-struct pppcp_data *pppcp_new(GAtPPP *ppp, const struct pppcp_proto *proto)
+struct pppcp_data *pppcp_new(GAtPPP *ppp, const struct pppcp_proto *proto,
+				gboolean dormant, guint max_failure)
 {
 	struct pppcp_data *data;
 
 	data = g_try_malloc0(sizeof(struct pppcp_data));
-	if (!data)
+	if (data == NULL)
 		return NULL;
 
-	data->state = INITIAL;
+	if (dormant)
+		data->state = STOPPED;
+	else
+		data->state = INITIAL;
+
 	data->config_timer_data.restart_interval = INITIAL_RESTART_TIMEOUT;
 	data->terminate_timer_data.restart_interval = INITIAL_RESTART_TIMEOUT;
 	data->config_timer_data.max_counter = MAX_CONFIGURE;
 	data->terminate_timer_data.max_counter = MAX_TERMINATE;
 	data->config_timer_data.data = data;
 	data->terminate_timer_data.data = data;
-	data->max_failure = MAX_FAILURE;
+
+	if (max_failure)
+		data->max_failure = max_failure;
+	else
+		data->max_failure = MAX_FAILURE;
 
 	data->ppp = ppp;
 	data->driver = proto;
